/*-------------------------------------------------------------------------
 *
 * pgsql_fdw.c
 *		  foreign-data wrapper for remote PostgreSQL servers.
 *
 * Copyright (c) 2012, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		  contrib/pgsql_fdw/pgsql_fdw.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"
#include "fmgr.h"

#include "catalog/pg_foreign_server.h"
#include "catalog/pg_foreign_table.h"
#include "commands/defrem.h"
#include "commands/explain.h"
#include "commands/vacuum.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "optimizer/cost.h"
#include "optimizer/pathnode.h"
#include "optimizer/planmain.h"
#include "optimizer/restrictinfo.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/rel.h"

#include "pgsql_fdw.h"
#include "connection.h"

PG_MODULE_MAGIC;

/*
 * Cost to establish a connection.
 * XXX: should be configurable per server?
 */
#define CONNECTION_COSTS			100.0

/*
 * Cost to transfer 1 byte from remote server.
 * XXX: should be configurable per server?
 */
#define TRANSFER_COSTS_PER_BYTE		0.001

/*
 * Describe the attribute where data conversion fails.
 */
typedef struct ErrorPos {
	Oid			relid;			/* oid of the foreign table */
	AttrNumber	cur_attno;		/* attribute number corrently processing */
} ErrorPos;

/*
 * FDW-specific information for RelOptInfo.fdw_private.  This is used to pass
 * information from pgsqlGetForeignRelSize to pgsqlGetForeignPaths.
 */
typedef struct PgsqlFdwPlanState {
	/*
	 * These are generated in GetForeignRelSize, and also used in subsequent
	 * GetForeignPaths.
	 */
	StringInfoData	sql;
	bool			has_where;
	Cost			startup_cost;
	Cost			total_cost;
	List		   *remote_conds;
	List		   *param_conds;
	List		   *local_conds;

	/* Cached catalog information. */
	ForeignTable   *table;
	ForeignServer  *server;
} PgsqlFdwPlanState;

/*
 * Index of FDW-private information stored in fdw_private list.
 *
 * We store various information in ForeignScan.fdw_private to pass them beyond
 * the boundary between planner and executor.  Finally FdwPlan holds items
 * below:
 *
 * 1) plain SELECT statement
 *
 * These items are indexed with the enum FdwPrivateIndex, so an item
 * can be accessed directly via list_nth().  For example of SELECT statement:
 *      sql = list_nth(fdw_private, FdwPrivateSelectSql)
 */
enum FdwPrivateIndex {
	/* SQL statements */
	FdwPrivateSelectSql,

	/* # of elements stored in the list fdw_private */
	FdwPrivateNum,
};

/*
 * Describes an execution state of a foreign scan against a foreign table
 * using pgsql_fdw.
 */
typedef struct PgsqlFdwExecutionState
{
	List	   *fdw_private;	/* FDW-private information */

	/* for remote query execution */
	PGconn	   *conn;			/* connection for the scan */
	Oid		   *param_types;	/* type array of external parameter */
	const char **param_values;	/* value array of external parameter */

	/* for tuple generation. */
	AttrNumber	attnum;			/* # of non-dropped attribute */
	Datum	   *values;			/* column value buffer */
	bool	   *nulls;			/* column null indicator buffer */
	AttInMetadata *attinmeta;	/* attribute metadata */

	/* for storing result tuples */
	MemoryContext scan_cxt;		/* context for per-scan lifespan data */
	MemoryContext temp_cxt;		/* context for per-tuple temporary data */
	Tuplestorestate *tuples;	/* result of the scan */

	/* for error handling. */
	ErrorPos		errpos;
} PgsqlFdwExecutionState;

/*
 * Describes a state of analyze request for a foreign table.
 */
typedef struct PgsqlAnalyzeState
{
	/* variables for tuple creation */
	TupleDesc		tupdesc;
	AttInMetadata  *attinmeta;
	char		   *colbuf;
	int				colbuflen;
	Datum		   *values;
	bool		   *nulls;

	/* variables for sampling */
	HeapTuple	   *rows;			/* result buffer */
	int				targrows;		/* target # of sample rows */
	int				numrows;		/* # of samples collected */
	double			samplerows;		/* # of rows fetched */
	double			rowstoskip;		/* # of rows skipped before next sample */
	double			rstate;			/* random state */

	/* for storing result tuples */
	MemoryContext anlz_cxt;		/* context for per-analyze lifespan data */
	MemoryContext temp_cxt;		/* context for per-tuple temporary data */

	/* for error handling. */
	ErrorPos		errpos;
} PgsqlAnalyzeState;

/*
 * SQL functions
 */
extern Datum pgsql_fdw_handler(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(pgsql_fdw_handler);

/*
 * FDW callback routines
 */
static void pgsqlGetForeignRelSize(PlannerInfo *root,
								   RelOptInfo *baserel,
								   Oid foreigntableid);
static void pgsqlGetForeignPaths(PlannerInfo *root,
								 RelOptInfo *baserel,
								 Oid foreigntableid);
static ForeignScan *pgsqlGetForeignPlan(PlannerInfo *root,
										RelOptInfo *baserel,
										Oid foreigntableid,
										ForeignPath *best_path,
										List *tlist,
										List *scan_clauses);
static void pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es);
static void pgsqlBeginForeignScan(ForeignScanState *node, int eflags);
static TupleTableSlot *pgsqlIterateForeignScan(ForeignScanState *node);
static void pgsqlReScanForeignScan(ForeignScanState *node);
static void pgsqlEndForeignScan(ForeignScanState *node);

static void pgsqlAnalyzeForeignTable(Relation relation,
									 VacuumStmt *vacstmt,
									 int elevel);

/*
 * Helper functions
 */
static void get_remote_estimate(const char *sql,
								PGconn *conn,
								double *rows,
								int *width,
								Cost *startup_cost,
								Cost *total_cost);
static void adjust_costs(double rows, int width,
						 Cost *startup_cost, Cost *total_cost);
static void execute_query(ForeignScanState *node);
static int query_row_processor(PGresult *res, const PGdataValue *columns,
							   const char **errmsgp, void *param);
static void pgsql_fdw_error_callback(void *arg);
static int pgsql_acquire_sample_rows(Relation onerel, HeapTuple *rows,
									 int targrows, double *totalrows,
									 double *totaldeadrows,
									 BlockNumber *totalpages, int elevel);
static int analyze_row_processor(PGresult *res, const PGdataValue *columns,
								 const char **errmsgp, void *param);

/* Exported functions, but not written in pgsql_fdw.h. */
void _PG_init(void);
void _PG_fini(void);

/*
 * Module-specific initialization.
 */
void
_PG_init(void)
{
}

/*
 * Module-specific clean up.
 */
void
_PG_fini(void)
{
}

/*
 * The content of FdwRoutine of pgsql_fdw is never changed, so we use one
 * static object for all scan.
 */
static FdwRoutine fdwroutine = {

	/* Node type */
	T_FdwRoutine,

	/* Required handler functions. */
	pgsqlGetForeignRelSize,
	pgsqlGetForeignPaths,
	pgsqlGetForeignPlan,
	pgsqlExplainForeignScan,
	pgsqlBeginForeignScan,
	pgsqlIterateForeignScan,
	pgsqlReScanForeignScan,
	pgsqlEndForeignScan,

	/* Optional handler functions. */
	pgsqlAnalyzeForeignTable,
};

/*
 * Foreign-data wrapper handler function: return a struct with pointers
 * to my callback routines.
 */
Datum
pgsql_fdw_handler(PG_FUNCTION_ARGS)
{
	PG_RETURN_POINTER(&fdwroutine);
}

/*
 * pgsqlGetForeignRelSize
 *		Estimate # of rows and width of the result of the scan
 *
 * Here we estimate number of rows returned by the scan in two steps.  In the
 * first step, we execute remote EXPLAIN command to obtain the number of rows
 * returned from remote side.  In the second step, we calculate the selectivity
 * of the filtering done on local side, and modify first estimate.
 *
 * We have to get some catalog objects and generate remote query string here,
 * so we store such expensive information in FDW private area of RelOptInfo and
 * pass them to subsequent functions for reuse.
 */
static void
pgsqlGetForeignRelSize(PlannerInfo *root,
					   RelOptInfo *baserel,
					   Oid foreigntableid)
{
	PgsqlFdwPlanState *fpstate;
	StringInfo		sql;
	ForeignTable   *table;
	ForeignServer  *server;
	UserMapping	   *user;
	PGconn		   *conn;
	double			rows;
	int				width;
	Cost			startup_cost;
	Cost			total_cost;
	List		   *remote_conds = NIL;
	List		   *param_conds = NIL;
	List		   *local_conds = NIL;
	Selectivity		sel;

	/*
	 * We use PgsqlFdwPlanState to pass various information to subsequent
	 * functions.
	 */
	fpstate = palloc0(sizeof(PgsqlFdwPlanState));
	initStringInfo(&fpstate->sql);
	sql = &fpstate->sql;
	fpstate->has_where = false;

	/* Retrieve catalog objects which are necessary to estimate rows. */
	table = GetForeignTable(foreigntableid);
	server = GetForeignServer(table->serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);

	/*
	 * Construct remote query which consists of SELECT, FROM, and WHERE
	 * clauses, but conditions contain any Param node are excluded because
	 * placeholder can't be used in EXPLAIN statement.  Such conditions are
	 * appended later.
	 */
	sortConditions(root, baserel, &remote_conds, &param_conds, &local_conds);
	deparseSimpleSql(sql, foreigntableid, root, baserel);
	if (list_length(remote_conds) > 0)
	{
		appendWhereClause(sql, fpstate->has_where, remote_conds, root);
		fpstate->has_where = true;
	}
	conn = GetConnection(server, user, false);
	get_remote_estimate(sql->data, conn, &rows, &width,
						&startup_cost, &total_cost);
	ReleaseConnection(conn);
	if (list_length(param_conds) > 0)
	{
		appendWhereClause(sql, fpstate->has_where, param_conds, root);
		fpstate->has_where = true;
	}

	/*
	 * Estimate selectivity of conditions which are not used in remote EXPLAIN
	 * by calling clauselist_selectivity().  The best we can do for
	 * parameterized condition is to estimate selectivity on the basis of local
	 * statistics.  When we actually obtain result rows, such conditions are
	 * deparsed into remote query and reduce rows transferred.
	 */
	sel = 1.0;
	sel *= clauselist_selectivity(root, param_conds,
								  baserel->relid, JOIN_INNER, NULL);
	sel *= clauselist_selectivity(root, local_conds,
								  baserel->relid, JOIN_INNER, NULL);
	baserel->rows = rows * sel;
	baserel->width = width;

	/*
	 * Pack obtained information into a object and store it in FDW-private area
	 * of RelOptInfo to pass them to subsequent functions.
	 */
	fpstate->startup_cost = startup_cost;
	fpstate->total_cost = total_cost;
	fpstate->remote_conds = remote_conds;
	fpstate->param_conds = param_conds;
	fpstate->local_conds = local_conds;
	fpstate->table = table;
	fpstate->server = server;
	baserel->fdw_private = (void *) fpstate;
}

/*
 * pgsqlGetForeignPaths
 *		Create possible scan paths for a scan on the foreign table
 */
static void
pgsqlGetForeignPaths(PlannerInfo *root,
					 RelOptInfo *baserel,
					 Oid foreigntableid)
{
	PgsqlFdwPlanState *fpstate = (PgsqlFdwPlanState *) baserel->fdw_private;
	ForeignPath	   *path;
	Cost			startup_cost;
	Cost			total_cost;
	List		   *fdw_private;

	/*
	 * We have cost values which are estimated on remote side, so use them to
	 * estimate better costs which respect various stuffs to complete the scan,
	 * such as sending query, transferring result, and local filtering.
	 *
	 * XXX We assume that remote cost factors are same as local, but it might
	 * be worth to make configurable.
	 */
	startup_cost = fpstate->startup_cost;
	total_cost = fpstate->total_cost;
	adjust_costs(baserel->rows, baserel->width, &startup_cost, &total_cost);

	/* Construct list of SQL statements and bind it with the path. */
	fdw_private = lappend(NIL, makeString(fpstate->sql.data));

	/*
	 * Create simplest ForeignScan path node and add it to baserel.  This path
	 * corresponds to SeqScan path of regular tables.
	 */
	path = create_foreignscan_path(root, baserel,
								   baserel->rows,
								   startup_cost,
								   total_cost,
								   NIL,				/* no pathkeys */
								   NULL,			/* no outer rel either */
								   NIL,				/* no param clause */
								   fdw_private);
	add_path(baserel, (Path *) path); 

	/*
	 * XXX We can consider sorted path or parameterized path here if we know
	 * that foreign table is indexed on remote end.  For this purpose, we
	 * might have to support FOREIGN INDEX to represent possible sets of sort
	 * keys and/or filtering.
	 */
}

/*
 * pgsqlGetForeignPlan
 *		Create ForeignScan plan node which implements selected best path
 */
static ForeignScan *
pgsqlGetForeignPlan(PlannerInfo *root,
					RelOptInfo *baserel,
					Oid foreigntableid,
					ForeignPath *best_path,
					List *tlist,
					List *scan_clauses)
{
	PgsqlFdwPlanState *fpstate = (PgsqlFdwPlanState *) baserel->fdw_private;
	Index			scan_relid = baserel->relid;
	List		   *fdw_private = NIL;
	List		   *fdw_exprs = NIL;
	List		   *local_exprs = NIL;
	ListCell	   *lc;

	/*
	 * We need lists of Expr other than the lists of RestrictInfo.  Now we can
	 * merge remote_conds and param_conds into fdw_exprs, because they are
	 * evaluated on remote side for actual remote query.
	 */
	foreach(lc, fpstate->remote_conds)
		fdw_exprs = lappend(fdw_exprs, ((RestrictInfo *) lfirst(lc))->clause);
	foreach(lc, fpstate->param_conds)
		fdw_exprs = lappend(fdw_exprs, ((RestrictInfo *) lfirst(lc))->clause);
	foreach(lc, fpstate->local_conds)
		local_exprs = lappend(local_exprs,
							  ((RestrictInfo *) lfirst(lc))->clause);

	/*
	 * Make a list contains SELECT statement to it to executor with plan node
	 * for later use.
	 */
	fdw_private = lappend(fdw_private, makeString(fpstate->sql.data));

	/*
	 * Create the ForeignScan node from target list, local filtering
	 * expressions, remote filtering expressions, and FDW private information.
	 *
	 * We remove expressions which are evaluated on remote side from qual of
	 * the scan node to avoid redundant filtering.  Such filter reduction
	 * can be done only here, done after choosing best path, because
	 * baserestrictinfo in RelOptInfo is shared by all possible paths until
	 * best path is chosen.
	 */
	return make_foreignscan(tlist,
							local_exprs,
							scan_relid,
							fdw_exprs,
							fdw_private);
}

/*
 * pgsqlExplainForeignScan
 *		Produce extra output for EXPLAIN
 */
static void
pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es)
{
	List	   *fdw_private;
	char	   *sql;

	fdw_private = ((ForeignScan *) node->ss.ps.plan)->fdw_private;
	sql = strVal(list_nth(fdw_private, FdwPrivateSelectSql));
	ExplainPropertyText("Remote SQL", sql, es);
}

/*
 * pgsqlBeginForeignScan
 *		Initiate access to a foreign PostgreSQL table.
 */
static void
pgsqlBeginForeignScan(ForeignScanState *node, int eflags)
{
	PgsqlFdwExecutionState *festate;
	PGconn		   *conn;
	Oid				relid;
	ForeignTable   *table;
	ForeignServer  *server;
	UserMapping	   *user;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;

	/*
	 * Do nothing in EXPLAIN (no ANALYZE) case.  node->fdw_state stays NULL.
	 */
	if (eflags & EXEC_FLAG_EXPLAIN_ONLY)
		return;

	/*
	 * Save state in node->fdw_state.
	 */
	festate = (PgsqlFdwExecutionState *) palloc(sizeof(PgsqlFdwExecutionState));
	festate->fdw_private = ((ForeignScan *) node->ss.ps.plan)->fdw_private;

	/*
	 * Create contexts for per-scan tuplestore under per-query context.
	 */
	festate->scan_cxt = AllocSetContextCreate(node->ss.ps.state->es_query_cxt,
											  "pgsql_fdw per-scan data",
											  ALLOCSET_DEFAULT_MINSIZE,
											  ALLOCSET_DEFAULT_INITSIZE,
											  ALLOCSET_DEFAULT_MAXSIZE);
	festate->temp_cxt = AllocSetContextCreate(node->ss.ps.state->es_query_cxt,
											  "pgsql_fdw temporary data",
											  ALLOCSET_DEFAULT_MINSIZE,
											  ALLOCSET_DEFAULT_INITSIZE,
											  ALLOCSET_DEFAULT_MAXSIZE);

	/*
	 * Get connection to the foreign server.  Connection manager would
	 * establish new connection if necessary.
	 */
	relid = RelationGetRelid(node->ss.ss_currentRelation);
	table = GetForeignTable(relid);
	server = GetForeignServer(table->serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);
	conn = GetConnection(server, user, true);
	festate->conn = conn;

	/* Result will be filled in first Iterate call. */
	festate->tuples = NULL;

	/* Allocate buffers for column values. */
	{
		TupleDesc	tupdesc = slot->tts_tupleDescriptor;
		festate->values = palloc(sizeof(Datum) * tupdesc->natts);
		festate->nulls = palloc(sizeof(bool) * tupdesc->natts);
		festate->attinmeta = TupleDescGetAttInMetadata(tupdesc);
	}

	/* Allocate buffers for query parameters. */
	{
		ParamListInfo	params = node->ss.ps.state->es_param_list_info;
		int				numParams = params ? params->numParams : 0;

		if (numParams > 0)
		{
			festate->param_types = palloc0(sizeof(Oid) * numParams);
			festate->param_values = palloc0(sizeof(char *) * numParams);
		}
		else
		{
			festate->param_types = NULL;
			festate->param_values = NULL;
		}
	}

	/* Remember which foreign table we are scanning. */
	festate->errpos.relid = relid;

	/* Store FDW-specific state into ForeignScanState */
	node->fdw_state = (void *) festate;

	return;
}

/*
 * pgsqlIterateForeignScan
 *		Retrieve next row from the result set, or clear tuple slot to indicate
 *		EOF.
 *
 *		Note that using per-query context when retrieving tuples from
 *		tuplestore to ensure that returned tuples can survive until next
 *		iteration because the tuple is released implicitly via ExecClearTuple.
 *		Retrieving a tuple from tuplestore in CurrentMemoryContext (it's a
 *		per-tuple context), ExecClearTuple will free dangling pointer.
 */
static TupleTableSlot *
pgsqlIterateForeignScan(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;
	MemoryContext	oldcontext = CurrentMemoryContext;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/*
	 * If this is the first call after Begin or ReScan, we need to execute
	 * remote query and get result set.
	 */
	if (festate->tuples == NULL)
		execute_query(node);

	/*
	 * If tuples are still left in tuplestore, just return next tuple from it.
	 *
	 * It is necessary to switch to per-scan context to make returned tuple
	 * valid until next IterateForeignScan call, because it will be released
	 * with ExecClearTuple then.  Otherwise, picked tuple is allocated in
	 * per-tuple context, and double-free of that tuple might happen.
	 * 
	 * If we don't have any result in tuplestore, clear result slot to tell
	 * executor that this scan is over.
	 */
	MemoryContextSwitchTo(festate->scan_cxt);
	tuplestore_gettupleslot(festate->tuples, true, false, slot);
	MemoryContextSwitchTo(oldcontext);

	return slot;
}

/*
 * pgsqlReScanForeignScan
 *   - Restart this scan by clearing old results and set re-execute flag.
 */
static void
pgsqlReScanForeignScan(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* If we haven't have valid result yet, nothing to do. */
	if (festate->tuples == NULL)
		return;

	/*
	 * Only rewind the current result set is enough.
	 */
	tuplestore_rescan(festate->tuples);
}

/*
 * pgsqlEndForeignScan
 *		Finish scanning foreign table and dispose objects used for this scan
 */
static void
pgsqlEndForeignScan(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* if festate is NULL, we are in EXPLAIN; nothing to do */
	if (festate == NULL)
		return;

	/*
	 * The connection which was used for this scan should be valid until the
	 * end of the scan to make the lifespan of remote transaction same as the
	 * local query.
	 */
	ReleaseConnection(festate->conn);
	festate->conn = NULL;

	/* Discard fetch results */
	if (festate->tuples != NULL)
	{
		tuplestore_end(festate->tuples);
		festate->tuples = NULL;
	}

	/* MemoryContext will be deleted automatically. */
}

/*
 * Collect statistics of a foreign table, and store the result in system
 * catalogs.
 */
static void
pgsqlAnalyzeForeignTable(Relation relation, VacuumStmt *vacstmt, int elevel)
{
	do_analyze_rel(relation, vacstmt, elevel, false, pgsql_acquire_sample_rows);
}

/*
 * Estimate costs of executing given SQL statement.
 */
static void
get_remote_estimate(const char *sql, PGconn *conn,
					double *rows, int *width,
					Cost *startup_cost, Cost *total_cost)
{
	PGresult	   *res = NULL;
	StringInfoData  buf;
	char		   *plan;
	char		   *p;
	int				n;

	/*
	 * Construct EXPLAIN statement with given SQL statement.
	 */
	initStringInfo(&buf);
	appendStringInfo(&buf, "EXPLAIN %s", sql);

	/* PGresult must be released before leaving this function. */
	PG_TRY();
	{
		res = PQexec(conn, buf.data);
		if (PQresultStatus(res) != PGRES_TUPLES_OK || PQntuples(res) == 0)
		{
			char *msg;

			msg = pstrdup(PQerrorMessage(conn));
			ereport(ERROR,
					(errmsg("could not execute EXPLAIN for cost estimation"),
					 errdetail("%s", msg),
					 errhint("%s", sql)));
		}

		/*
		 * Find estimation portion from top plan node. Here we search opening
		 * parentheses from the end of the line to avoid finding unexpected
		 * parentheses.
		 */
		plan = PQgetvalue(res, 0, 0);
		p = strrchr(plan, '(');
		if (p == NULL)
			elog(ERROR, "wrong EXPLAIN output: %s", plan);
		n = sscanf(p,
				   "(cost=%lf..%lf rows=%lf width=%d)",
				   startup_cost, total_cost, rows, width);
		if (n != 4)
			elog(ERROR, "could not get estimation from EXPLAIN output");

		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();
}

/*
 * Adjust costs estimated on remote end with some overheads such as connection
 * and data transfer.
 */
static void
adjust_costs(double rows, int width, Cost *startup_cost, Cost *total_cost)
{
	/*
	 * TODO Selectivity of quals which are NOT pushed down should be also
	 * considered.
	 */

	/* add cost to establish connection. */
	*startup_cost += CONNECTION_COSTS;
	*total_cost += CONNECTION_COSTS;

	/* add cost to transfer result. */
	*total_cost += TRANSFER_COSTS_PER_BYTE * width * rows;
	*total_cost += cpu_tuple_cost * rows;
}

/*
 * Execute remote query with current parameters.
 */
static void
execute_query(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	ParamListInfo	params = node->ss.ps.state->es_param_list_info;
	int				numParams = params ? params->numParams : 0;
	Oid			   *types = NULL;
	const char	  **values = NULL;
	char		   *sql;
	PGconn		   *conn;
	PGresult	   *res = NULL;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;
	types = festate->param_types;
	values = festate->param_values;

	/*
	 * Construct parameter array in text format.  We don't release memory for
	 * the arrays explicitly, because the memory usage would not be very large,
	 * and anyway they will be released in context cleanup.
	 */
	if (numParams > 0)
	{
		int i;

		for (i = 0; i < numParams; i++)
		{
			types[i] = params->params[i].ptype;
			if (params->params[i].isnull)
				values[i] = NULL;
			else
			{
				Oid			out_func_oid;	
				bool		isvarlena;
				FmgrInfo	func;

				getTypeOutputInfo(types[i], &out_func_oid, &isvarlena);
				fmgr_info(out_func_oid, &func);
				values[i] = OutputFunctionCall(&func, params->params[i].value);
			}
		}
	}

	/* PGresult must be released before leaving this function. */
	PG_TRY();
	{
		/*
		 * Execute remote query with parameters, and retrieve results with
		 * custom row processor which stores results in tuplestore.
		 *
		 * We uninstall the custom row processor right after processing all
		 * results.
		 */
		conn = festate->conn;
		sql = strVal(list_nth(festate->fdw_private, FdwPrivateSelectSql));
		PQsetRowProcessor(conn, query_row_processor, node);
		res = PQexecParams(conn, sql, numParams, types, values, NULL, NULL, 0);
		PQsetRowProcessor(conn, NULL, NULL);

		/*
		 * We can't know whether the scan is over or not in custom row
		 * processor, so mark that the result is valid here.
		 */
		tuplestore_donestoring(festate->tuples);

		/*
		 * If the query has failed, reporting details is enough here.
		 * Connection(s) which are used by this query (at least used by
		 * pgsql_fdw) will be cleaned up by the foreign connection manager.
		 */
		if (PQresultStatus(res) != PGRES_TUPLES_OK)
		{
			ereport(ERROR,
					(errmsg("could not execute remote query"),
					 errdetail("%s", PQerrorMessage(conn)),
					 errhint("%s", sql)));
		}

		/* Discard result of SELECT statement. */
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();
}

/*
 * Create tuples from PGresult and store them into tuplestore.
 *
 * Caller must use PG_TRY block to catch exception and release PGresult
 * surely.
 */
static int
query_row_processor(PGresult *res,
					const PGdataValue *columns,
					const char **errmsgp,
					void *param)
{
	ForeignScanState   *node = (ForeignScanState *) param;
	int			nfields = PQnfields(res);
	int			i;
	int			j;
	int			attnum;		/* number of non-dropped columns */
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;
	TupleDesc	tupdesc = slot->tts_tupleDescriptor;
	Form_pg_attribute  *attrs = tupdesc->attrs;
	PgsqlFdwExecutionState *festate = (PgsqlFdwExecutionState *) node->fdw_state;
	AttInMetadata *attinmeta = festate->attinmeta;
	char	   *colbuf;
	int			colbuflen;
	HeapTuple	tuple;
	ErrorContextCallback errcontext;
	MemoryContext oldcontext;

	if (columns == NULL)
	{
		/* count non-dropped columns */
		for (attnum = 0, i = 0; i < tupdesc->natts; i++)
			if (!attrs[i]->attisdropped)
				attnum++;

		/* check result and tuple descriptor have the same number of columns */
		if (attnum > 0 && attnum != nfields)
			ereport(ERROR,
					(errcode(ERRCODE_DATATYPE_MISMATCH),
					 errmsg("remote query result rowtype does not match "
							"the specified FROM clause rowtype"),
					 errdetail("expected %d, actual %d", attnum, nfields)));

		/* First, ensure that the tuplestore is empty. */
		if (festate->tuples == NULL)
		{

			/*
			 * Create tuplestore to store result of the query in per-query
			 * context.  Note that we use this memory context to avoid memory
			 * leak in error cases.
			 */
			oldcontext = MemoryContextSwitchTo(festate->scan_cxt);
			festate->tuples = tuplestore_begin_heap(false, false, work_mem);
			MemoryContextSwitchTo(oldcontext);
		}
		else
		{
			/* Clear old result just in case. */
			tuplestore_clear(festate->tuples);
		}

		return 1;
	}

	/*
	 * This function is called repeatedly until all result rows are processed,
	 * so we should allow interrupt.
	 */
	CHECK_FOR_INTERRUPTS();

	/*
	 * Do the following work in a temp context that we reset after each tuple.
	 * This cleans up not only the data we have direct access to, but any
	 * cruft the I/O functions might leak.
	 */
	oldcontext = MemoryContextSwitchTo(festate->temp_cxt);

	/*
	 * Set up and install callback to identify error column.  We don't set
	 * callback right row because it should be set only during column value
	 * conversion.
	 */
	errcontext.callback = pgsql_fdw_error_callback;
	errcontext.arg = (void *) &festate->errpos;
	errcontext.previous = error_context_stack;
	error_context_stack = &errcontext;

	/* Initialize column value buffer. */
	colbuflen = 1024;
	colbuf = palloc(colbuflen);

	for (i = 0, j = 0; i < tupdesc->natts; i++)
	{
		int		len = columns[j].len;

		/* skip dropped columns. */
		if (attrs[i]->attisdropped)
		{
			festate->nulls[i] = true;
			continue;
		}

		/*
		 * Set NULL indicator, and convert text representation to internal
		 * representation if any.
		 */
		if (len < 0)
			festate->nulls[i] = true;
		else
		{
			Datum	value;

			while (colbuflen < len + 1)
			{
				colbuflen *= 2;
				colbuf = repalloc(colbuf, colbuflen);
			}
			memcpy(colbuf, columns[j].value, len);
			colbuf[columns[j].len] = '\0';

			festate->errpos.cur_attno = i + 1;
			festate->nulls[i] = false;
			value = InputFunctionCall(&attinmeta->attinfuncs[i],
									  colbuf,
									  attinmeta->attioparams[i],
									  attinmeta->atttypmods[i]);
			festate->values[i] = value;
		}
		j++;
	}

	/* Uninstall error callback function. */
	error_context_stack = errcontext.previous;

	/*
	 * Build the tuple and put it into the slot.
	 * We don't have to free the tuple explicitly because it's been
	 * allocated in the per-tuple context.
	 */
	tuple = heap_form_tuple(tupdesc, festate->values, festate->nulls);
	tuplestore_puttuple(festate->tuples, tuple);

	/* Clean up */
	MemoryContextSwitchTo(oldcontext);
	MemoryContextReset(festate->temp_cxt);

	return 1;
}

/*
 * Callback function which is called when error occurs during column value
 * conversion.  Print names of column and relation.
 */
static void
pgsql_fdw_error_callback(void *arg)
{
	ErrorPos *errpos = (ErrorPos *) arg;
	const char	   *relname;
	const char	   *colname;

	relname = get_rel_name(errpos->relid);
	if (errpos->cur_attno != InvalidAttrNumber)
	{
		colname = get_attname(errpos->relid, errpos->cur_attno);
		errcontext("column %s of foreign table %s",
				   quote_identifier(colname), quote_identifier(relname));
	}
	else
		errcontext("foreign table %s", quote_identifier(relname));
}

/*
 * Acquire a random sample of rows from foreign table managed by pgsql_fdw.
 *
 * pgsql_fdw doesn't provide direct access to remote buffer, so we exeucte
 * simple SELECT statement which retrieves whole rows from remote side, and
 * pick some samples from them.
 */
static int
pgsql_acquire_sample_rows(Relation onerel, HeapTuple *rows, int targrows,
						  double *totalrows, double *totaldeadrows,
						  BlockNumber *totalpages, int elevel)
{
	PgsqlAnalyzeState astate;
	StringInfoData sql;
	ForeignTable *table;
	ForeignServer *server;
	UserMapping *user;
	PGconn	   *conn = NULL;
	PGresult   *res = NULL;

	/*
	 * Only few information are necessary as input to row processor.  Other
	 * initialization will be done at the first row processor call.
	 */
	astate.anlz_cxt = CurrentMemoryContext;
	astate.temp_cxt = AllocSetContextCreate(CurrentMemoryContext,
											"pgsql_fdw analyze temporary data",
											ALLOCSET_DEFAULT_MINSIZE,
											ALLOCSET_DEFAULT_INITSIZE,
											ALLOCSET_DEFAULT_MAXSIZE);
	astate.rows = rows;
	astate.targrows = targrows;
	astate.tupdesc = onerel->rd_att;
	astate.errpos.relid = onerel->rd_id;

	/*
	 * Construct SELECT statement which retrieves whole rows from remote.  We
	 * can't avoid running sequential scan on remote side to get practical
	 * statistics, so this seems reasonable compromise.
	 */
	initStringInfo(&sql);
	deparseAnalyzeSql(&sql, onerel);

	table = GetForeignTable(onerel->rd_id);
	server = GetForeignServer(table->serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);
	conn = GetConnection(server, user, true);

	/*
	 * Acquire sample rows from the result set.
	 */
	PG_TRY();
	{
		/* Declare non-scrollable cursor for analyze. */
		PQsetRowProcessor(conn, analyze_row_processor, &astate);
		res = PQexec(conn, sql.data);
		PQsetRowProcessor(conn, NULL, NULL);
		if (PQresultStatus(res) != PGRES_TUPLES_OK)
			ereport(ERROR,
					(errmsg("could not execute remote query for analyze"),
					 errdetail("%s", PQerrorMessage(conn)),
					 errhint("%s", sql.data)));
		PQclear(res);
		res = NULL;
	}
	PG_CATCH();
	{
		PQclear(res);
		PG_RE_THROW();
	}
	PG_END_TRY();

	ReleaseConnection(conn);

	/* We assume that we have no dead tuple. */
	*totaldeadrows = 0.0;

	/* We've retrieved all living tuples from foreign server. */
	*totalrows = astate.samplerows;

	/*
	 * We don't update pg_class.relpages because we don't care that in
	 * planning at all.
	 */

	/*
	 * Emit some interesting relation info
	 */
	ereport(elevel,
			(errmsg("\"%s\": scanned with \"%s\", "
					"containing %.0f live rows and %.0f dead rows; "
					"%d rows in sample, %.0f estimated total rows",
					RelationGetRelationName(onerel), sql.data,
					astate.samplerows, 0.0,
					astate.numrows, astate.samplerows)));

	return astate.numrows;
}

/*
 * Custom row processor for acquire_sample_rows.
 *
 * Collect sample rows from the result of query.
 *   - Use all tuples as sample until target rows samples are collected. 
 *   - Once reached the target, skip some tuples and replace already sampled
 *     tuple randomly.
 */
static int
analyze_row_processor(PGresult *res, const PGdataValue *columns,
					  const char **errmsgp, void *param)
{
	PgsqlAnalyzeState *astate = (PgsqlAnalyzeState *) param;
	int			targrows = astate->targrows;
	TupleDesc	tupdesc = astate->tupdesc;
	AttInMetadata *attinmeta = astate->attinmeta;
	int			i;
	int			j;
	int			pos;	/* position where next sample should be stored. */
	HeapTuple	tuple;
	ErrorContextCallback errcontext;
	MemoryContext callercontext;

	if (columns == NULL)
	{
		/* Prepare for sampling rows */
		astate->attinmeta = TupleDescGetAttInMetadata(tupdesc);
		astate->colbuflen = 1024;
		astate->colbuf = palloc(astate->colbuflen);
		astate->values = (Datum *) palloc(sizeof(Datum) * tupdesc->natts);
		astate->nulls = (bool *) palloc(sizeof(bool) * tupdesc->natts);
		astate->numrows = 0;
		astate->samplerows = 0;
		astate->rowstoskip = -1;
		astate->numrows = 0;
		astate->rstate = init_selection_state(astate->targrows);

		return 1;
	}

	/*
	 * ANALYZE against foreign tables are not done in processes of
	 * vacuum, so here we use CHECK_FOR_INTERRUPTS instead of
	 * vacuum_delay_point().
	 */
	CHECK_FOR_INTERRUPTS();

	/*
	 * Do the following work in a temp context that we reset after each tuple.
	 * This cleans up not only the data we have direct access to, but any
	 * cruft the I/O functions might leak.
	 */
	callercontext = MemoryContextSwitchTo(astate->temp_cxt);

	/*
	 * Set up and install callback to identify error column.  We don't set
	 * callback right row because it should be set only during column value
	 * conversion.  However, we need to use the context which has loger
	 * lifespan for column buffer and result tuple.
	 */
	errcontext.callback = pgsql_fdw_error_callback;
	errcontext.arg = (void *) &astate->errpos;
	errcontext.previous = error_context_stack;
	error_context_stack = &errcontext;

	/*
	 * First targrows rows are once sampled always.  If we have more source
	 * rows, pick up some of them by skipping and replace already sampled
	 * tuple randomly.
	 *
	 * Here we just deterime the slot where next sample should be stored.  Set
	 * pos to negative value to indicates the row should be skipped.
	 */
	if (astate->numrows < targrows)
		pos = astate->numrows++;
	else
	{
		/*
		 * The first targrows sample rows are simply copied into
		 * the reservoir.  Then we start replacing tuples in the
		 * sample until we reach the end of the relation.  This
		 * algorithm is from Jeff Vitter's paper, similarly to
		 * acquire_sample_rows in analyze.c.
		 *
		 * We don't have block-wise accessibility, so every row in
		 * the PGresult is possible to be sample.
		 */
		if (astate->rowstoskip < 0)
			astate->rowstoskip = get_next_S(astate->samplerows, targrows,
											&astate->rstate);

		if (astate->rowstoskip <= 0)
		{
			int		k = (int) (targrows * random_fract());

			Assert(k >= 0 && k < targrows);

			/*
			 * Create sample tuple from the result, and replace at
			 * random.
			 */
			heap_freetuple(astate->rows[k]);
			pos = k;
		}
		else
			pos = -1;

		astate->rowstoskip -= 1;
	}

	/* Always increment sample row counter. */
	astate->samplerows += 1;

	if (pos >= 0)
	{
		/*
		 * Create sample tuple from current result row, and store it into the
		 * position determined above.  Note that i and j point entries in
		 * catalog and columns array respectively.
		 */
		for (i = 0, j = 0; i < tupdesc->natts; i++)
		{
			int		len = columns[j].len;

			if (tupdesc->attrs[i]->attisdropped)
				continue;

			if (len < 0)
				astate->nulls[i] = true;
			else
			{
				Datum	value;

				/*
				 * Expand column value string buffer twice as current until
				 * it can hold the value.  We need to allocate the buffer in
				 * the context of analyze.
				 */
				while (astate->colbuflen < len + 1)
				{
					astate->colbuflen *= 2;
					MemoryContextSwitchTo(astate->anlz_cxt);
					astate->colbuf = repalloc(astate->colbuf,
											  astate->colbuflen);
					MemoryContextSwitchTo(astate->temp_cxt);
				}
				memcpy(astate->colbuf, columns[j].value, len);
				astate->colbuf[columns[j].len] = '\0';

				astate->errpos.cur_attno = i + 1;
				astate->nulls[i] = false;
				value = InputFunctionCall(&attinmeta->attinfuncs[i],
										  astate->colbuf,
										  attinmeta->attioparams[i],
										  attinmeta->atttypmods[i]);
				astate->values[i] = value;
			}
			j++;
		}

		/*
		 * Generate tuple from the result row data, and store it into the give
		 * buffer.  Note that we need to allocate the tuple in the analyze
		 * context to make it valid even after temporary per-tuple context has
		 * been reset.
		 */
		MemoryContextSwitchTo(astate->anlz_cxt);
		tuple = heap_form_tuple(tupdesc, astate->values, astate->nulls);
		MemoryContextSwitchTo(astate->temp_cxt);
		astate->rows[pos] = tuple;
	}

	/* Uninstall error callback function. */
	error_context_stack = errcontext.previous;

	/* Clean up */
	MemoryContextSwitchTo(callercontext);
	MemoryContextReset(astate->temp_cxt);

	return 1;
}
